/*
** RasterColor
**
**	This module does the color hacking functions. This module is based
**	on the file 'xvcolor.c' from John Bradley's popular 'xv' program:
**
**      xvcolor.c - color allocation/sorting/freeing code
**	John Bradley, University of Pennsylvania (bradley@cis.upenn.edu)
**
** Copyright 1992, 1993 by Markku Savela and
**	Technical Research Centre of Finland
**
** The author may be contacted via:
**
**    Mail:	Markku Savela
**		Technical Research Centre of Finland
**		Telecommunications Laboratory
**		Otakaari 7 B, P.O. Box 34
**		SF-02151 ESPOO
**		Finland
**
**    Phone:	+358 0 456 6509
**    Email:	savela@tel.vtt.fi
*/

#if SYSV_INCLUDES
#	include <malloc.h>	/* malloc() et al. */
#	include <memory.h>	/* memset() et al. */
#endif
#if ANSI_INCLUDES
#	include <stdlib.h>
#	include <string.h>	/* memset() et al. */
#endif
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xew/RasterP.h>
#include "RasterColor.h"

/* MONO returns total intensity of r,g,b components */
#define MONO(rd,gn,bl) (((rd)*11 + (gn)*16 + (bl)*5) >> 5)  /*.33R+ .5G+ .17B*/

/*
** structure and routine used in SortColormap()
*/
typedef unsigned char byte;
typedef struct thing
    {
	byte r,g,b,n;		/* actual value of color + alignment */
	int oldindex;           /* its index in the old colormap */
	int use;                /* # of pixels of this color */
	int mindist;            /* min distance to a selected color */
    } CMAPENT;

static byte colAllocOrder[256];	/* color allocations order */

#if NeedFunctionPrototypes
static Status AllocColor(XeRasterWidget, Colormap, XColor *);
static void FreeColors(XeRasterWidget, Colormap, unsigned long *,
		       int, unsigned long);
#else
static Status AllocColor();
static void FreeColors();
#endif

static void SortColormap(w, r, g, b)
XeRasterWidget w;
byte *r, *g, *b;
    {
	byte *p, *q;
	byte *ir, *ig, *ib;

	int   i, j, mdist, entry, d, hist[256], trans[256];
	static CMAPENT c[256], c1[256], *cp, *cj, *ck;
	XImage *image = w->raster.image;
	int cMapSize = (1 << w->raster.bits_per_sample);

	if (image == 0)
		XeWidgetErrorMsg
			((Widget)w, "RasterColor",
			 "Color allocation called without XImage",
			 (String *)NULL,0);
	if (cMapSize > 256)
		XeWidgetErrorMsg
			((Widget)w, "RasterColor",
			 "Cannot handle larger than 8bit pixels",
			 (String *)NULL,0);
	ir = w->raster.color_map;
	ig = ir + cMapSize;
	ib = ig + cMapSize;

	/* initialize histogram and compute it */

	memset((void *)hist, 0, sizeof(hist));
	p = (byte *)image->data;
	for (i = image->height; i > 0; i--, p += image->bytes_per_line)
		for (j = image->width, q = p; j > 0; j--, q++)
			hist[*q]++;
	/*
	** Put the actually-used colors into the 'c' array in the order
	** they occur also, while we're at it, number of colors and close
	** up gaps in colortable
	*/
	for (i = w->raster.num_colors = 0, cp = c; i < cMapSize; i++) 
		if (hist[i]) 
		    { 
			r[w->raster.num_colors] = ir[i];
			g[w->raster.num_colors] = ig[i];
			b[w->raster.num_colors] = ib[i];
			trans[i] = w->raster.num_colors;
			cp->r = ir[i];
			cp->g = ig[i];
			cp->b = ib[i];
			cp->use = hist[i]; 
			cp->oldindex = w->raster.num_colors;
			cp->mindist = 200000; /* 255^2 * 3 = 195075 */
			w->raster.num_colors += 1;
			cp += 1;
		    }
	/*
	** modify image to reflect new (compressed, but not reordered)
	** colormap
	*/
	p = (byte *)image->data;
	for (i = image->height; i; i--, p += image->bytes_per_line)
		for (j = image->width, q = p; j; j--, q++)
			*q = trans[*q];
	/*
	** find most-used color, put that in c1[0]
	*/
	for (entry = -1, mdist = -1, i=0; i < w->raster.num_colors; i++)
		if (c[i].use > mdist)
		    {
			mdist = c[i].use; 
			entry=i; 
		    }
	memcpy((void *)&c1[0], (void *)&c[entry], sizeof(CMAPENT));
	c[entry].use = 0;   /* and mark it dealt with */
	/*
	** sort rest of colormap.  Half of the entries are allocated on the
	** basis of distance from already allocated colors, and half on the
	** basis of usage.  (NB: 'taxicab' distance is used throughout this
	** file.)
	**
	** Mod:  pick first 10 colors based on maximum distance.  pick
	** remaining colors half by distance and half by usage   -- JHB
	**
	** To obtain O(n^2) performance, we keep each unselected color
	** (in c[], with use>0) marked with the minimum distance to any of
	** the selected colors (in c1[]).  Each time we select a color, we
	** can update the minimum distances in O(n) time. 
	**
	** mod by Tom Lane   Tom.Lane@g.gp.cs.cmu.edu
	**
	** (above comment was in xvraster.c --msa)
	*/
	for (i = 1; i < w->raster.num_colors; i++)
	    {
		int ckR, ckG, ckB;
		/*
		** Get RGB of color last selected  and choose
		** selection method
		*/
		ck = &c1[i-1];            /* point to just-selected color */
		ckR = ck->r;
		ckG = ck->g;
		ckB = ck->b;
		entry = -1;
		mdist = -1;
		if (i&1 || i<10) 
			/*
			** Now find the i'th most different color.
			** we want to select the unused color that has
			** the greatest mindist
			*/
			for (j = 0,cj = c; j < w->raster.num_colors; j++,cj++)
			    {
				if (cj->use)
				    { 
					/* this color has not been marked
					   already. update mindist */
					d = (cj->r - ckR) * (cj->r - ckR) +
					    (cj->g - ckG) * (cj->g - ckG) + 
					    (cj->b - ckB) * (cj->b - ckB);
					if (cj->mindist > d)
						cj->mindist = d;
					if (cj->mindist > mdist)
					    {
						mdist = cj->mindist;
						entry = j;
					    }
				    }
			    }
		else 
			/* Now find the i'th most different color */
			/* we want to select the unused color that has
			   the greatest usage */
			for (j=0, cj=c; j < w->raster.num_colors; j++,cj++)
			    {
				if (cj->use)
				    {  /* this color has not been marked
					  already. update mindist */
					d = (cj->r - ckR) * (cj->r - ckR) +
					    (cj->g - ckG) * (cj->g - ckG) + 
					    (cj->b - ckB) * (cj->b - ckB);
					if (cj->mindist > d)
						cj->mindist = d;
					if (cj->use > mdist) 
					    {
						mdist = cj->use; 
						entry = j;
					    }
				    }
			    }
		/* c[entry] is the next color to put in the map.  do so */
		memcpy((void *)&c1[i], (void *)&c[entry], sizeof(CMAPENT));
		c[entry].use = 0;
	    }
	for ( i = 0; i < w->raster.num_colors; i++)
		colAllocOrder[i] = c1[i].oldindex;
    }

#define NOPIX 0xffffffff    

static Colormap SetupColormap(w, private)
XeRasterWidget w;
int private;
    {
	XeRasterFreeColors(w);
	w->raster.num_allocated = 0;
	w->raster.alloc =(unsigned long *)
		XtMalloc(w->raster.num_colors * sizeof(unsigned long));
	if (private)
	    {
		Display *display = XtDisplay(w);
		Colormap cmap;

		cmap = XCreateColormap(display,XtWindow(w),
				       w->raster.visual,AllocNone);
		if (cmap)
		    {
			w->raster.cmap = w->core.colormap;
			w->core.colormap = cmap;
			XSetWindowColormap(display, XtWindow(w), cmap);
		    }
	    }
	return w->core.colormap;
    }

/*
** XeRasterAllocColors
**	Remap and allocate colors for the Image stored in raster.image.
**	This function is only designed to be called with PseudoColor
**	and GrayScale visuals. Calling with others may produce incorrect
**	results.
*/
void XeRasterAllocColors(w)
XeRasterWidget w;
    {
	int i, j, c, unique, p2alloc, p3alloc;
	Colormap cmap;
	XColor defs[256];
	XColor ctab[256];
	unsigned long cols[256];/* map pic pixel values to X pixel vals */
	int fc2pcol[256];	/* maps freecols into pic pixel values */
	int dc;
	byte r[256], g[256], b[256], *p, *q;
	XImage *image = w->raster.image;
	int mono;

	switch (w->raster.visual->class)
	    {
	    case PseudoColor:	/* .. the primary support */
	    case StaticColor:	/* .. might work */
		mono = False;
		break;
	    case GrayScale:
	    case StaticGray:
	    default:		/* .. do mono for all others to be safe */
		mono = True;
	    }
	unique = p2alloc = p3alloc = 0;
	if (w->raster.max_colors == 0)
		/* Dither with 'black' & 'white' */
		return;

	SortColormap(w, r, g, b);
	cmap = SetupColormap
		(w, w->raster.colormap_use == XeColormapUse_PRIVATE);
	for (i=0; i < w->raster.num_colors; i++)
		cols[i] = NOPIX;
	/*
	** FIRST PASS COLOR ALLOCATION:  
	** for each color in the 'desired colormap', try to get it via
	** AllocColor().  If for any reason it fails, mark that pixel
	** 'unallocated' and worry about it later.  Repeat.
	**
	** attempt to allocate first max_colors entries in colormap 
	** note: On displays with less than 8 bits per RGB gun, it's quite
	** possible that different colors in the original picture will be
	** mapped to the same color on the screen.  X does this for you
	** silently.  However, this is not-desirable for this application, 
	** because when I say 'allocate me 32 colors' I want it to allocate
	** 32 different colors, not 32 instances of the same 4 shades...
	*/
	for (i=0; i<w->raster.num_colors && unique<w->raster.max_colors; i++)
	    {
		c = colAllocOrder[i];
		if (mono)
		    { 
			int intens = MONO(r[c], g[c], b[c]);
			defs[c].red = defs[c].green = defs[c].blue = intens<<8;
		    }
		else
		    {
			defs[c].red   = r[c]<<8;
			defs[c].green = g[c]<<8;
			defs[c].blue  = b[c]<<8;
		    }
		defs[c].flags = DoRed | DoGreen | DoBlue;
		if (AllocColor(w, cmap, &defs[c]))
		    { 
			unsigned long pixel, *fcptr;

			pixel = cols[c] = defs[c].pixel;
			/*
			** see if the newly allocated color is new and
			** different
			*/
			for (j=0, fcptr = w->raster.alloc;;)
				if (j++ == w->raster.num_allocated)
				    {
					unique++;
					break;
				    }
				else if (*fcptr++ == pixel)
					break;
			fc2pcol[w->raster.num_allocated] = c;
			w->raster.alloc[w->raster.num_allocated++] = pixel;
		    }
		else if (w->raster.colormap_use == XeColormapUse_OPTIONAL &&
			 !w->raster.cmap)
		    {
			/*
			** the allocation failed.  If we want 'perfect' color,
			** and we haven't already created our own colormap,
			** we'll want to do so
			*/
			cmap = SetupColormap(w, True);
			for (i=0; i < w->raster.num_colors; i++)
				cols[i] = NOPIX;
			unique = 0;
			i = -1;
		    }
		else
			/*
			** either we don't care about perfect color,
			** or we do care, have allocated our own
			** colormap, and have STILL run out of colors
			** (possible, even on an 8 bit display), just
			** mark pixel as unallocated.  We'll deal with
			** it later
			*/
			cols[c] = NOPIX;
	    }  /* FIRST PASS */

	if (w->raster.num_allocated == w->raster.num_colors)
		goto colorallocdone;
	/*
	** SECOND PASS COLOR ALLOCATION:
	** Allocating 'exact' colors failed.  Now try to allocate 'closest'
	** colors.
	** 
	** Read entire X colormap (or first 256 entries) in from display.
	** for each unallocated pixel, find the closest color that actually
	** is in the X colormap.  Try to allocate that color (read only).
	** If that fails, the THIRD PASS will deal with it
	*/
	
	/* read entire colormap (or first 256 entries) into 'ctab' */
	dc = (w->raster.visual->map_entries < 256) ? 
		w->raster.visual->map_entries : 256;
	for (i = 0; i < dc; i++)
		ctab[i].pixel = (unsigned long)i;
	XQueryColors(XtDisplay(w), cmap, ctab, dc);
	for (i=0; i < w->raster.num_colors && unique < w->raster.max_colors;
	     i++)
	    {
		int d, mdist, close;
		unsigned long ri,gi,bi;
			
		c = colAllocOrder[i];
		if (cols[c] != NOPIX)
			continue;

		/* an unallocated pixel */

		mdist = 100000;
		close = -1;
		ri = r[c]; 
		gi = g[c];
		bi = b[c];
		for (j = 0; j < dc; j++)
		    {
			d = abs(ri - (ctab[j].red   >> 8)) +
			    abs(gi - (ctab[j].green >> 8)) +
			    abs(bi - (ctab[j].blue  >> 8));
			if (d < mdist)
			    {
				mdist=d;
				close=j;
			    }
		    }
		if (close < 0)
			XeWidgetErrorMsg
				((Widget)w, "RasterColor","This Can't Happen!",
				 (String *)NULL,0);
		if (AllocColor(w, cmap, &ctab[close]))
		    { 
			memcpy((void *)&defs[c], (void*)&ctab[close],
			       sizeof(XColor));
			cols[c]  = ctab[close].pixel;
			fc2pcol[w->raster.num_allocated] = c;
			w->raster.alloc[w->raster.num_allocated++] = cols[c];
			p2alloc++;
			unique++;
		    }
	    }
	if (w->raster.num_allocated == w->raster.num_colors)
		goto colorallocdone;
	/*
	** THIRD PASS COLOR ALLOCATION:
	** We've alloc'ed all the colors we can.  Now, we have to map any
	** remaining unalloced pixels into either A) the colors that we DID get
	** (noglob), or B) the colors found in the X colormap
	*/
	for (i=0; i < w->raster.num_colors; i++)
	    {
		int d, k, mdist, close;
		unsigned long ri,gi,bi;

		c = colAllocOrder[i];
		if (cols[c] != NOPIX)
			continue;
		/* an unallocated pixel */
			
		mdist = 100000;   close = -1;
		ri = r[c];  gi = g[c];  bi = b[c];
		if (w->raster.colormap_use != XeColormapUse_SHAREDOWN)
		    {   /* search the entire X colormap */
			for (j = 0; j < dc; j++)
			    {
				d = abs(ri - (ctab[j].red   >> 8)) +
				    abs(gi - (ctab[j].green >> 8)) +
				    abs(bi - (ctab[j].blue  >> 8));
				if (d < mdist)
				    {
					mdist = d;
					close = j;
				    }
			    }
			p3alloc++;
		    }
		else
		    { 
			/* only search the alloc'd colors */
			for (j = 0; j < w->raster.num_allocated; j++)
			    {
				k = fc2pcol[j];
				d = abs(ri - (defs[k].red >> 8)) +
				    abs(gi - (defs[k].green >> 8)) +
				    abs(bi - (defs[k].blue >> 8));
				if (d < mdist)
				    { 
					mdist = d;
					close = k;
				    }
			    }
		    }
		if (close < 0)
			XeWidgetErrorMsg
				((Widget)w, "RasterColor",
				 "This Can't Happen!", (String *)NULL,0);
		memcpy((void *)&defs[c], (void *)&ctab[close], sizeof(XColor));
		cols[c]  = defs[c].pixel;
	    }  /* THIRD PASS */
    colorallocdone:
	/*
	** Change pixel values in the image to match the allocated
	** colors.
	*/
	p = (byte *)image->data;
	for (i = image->height; i > 0; i--, p += image->bytes_per_line)
		for (j = image->width, q = p; j > 0; j--, q++)
			*q = cols[*q];
    }

/*******************************************************/
/* 24/32-bit TrueColor display color 'allocation' code */
/*******************************************************/

static int highbit(ul)
unsigned long ul;
    {
	/* returns position of highest set bit in 'ul' as an integer (0-31),
	   or -1 if none */
	
	int i;
	for (i=31; ((ul&0x80000000) == 0) && i>=0;  i--, ul<<=1);
	return i;
    }


static Status AllocColor(w, cm, cdef)
XeRasterWidget w;
Colormap cm;
XColor *cdef;
    {
	Visual *visual = w->raster.visual;

	if (visual->class == TrueColor || visual->class == DirectColor)
	    {
		unsigned long r, g, b, rmask, gmask, bmask;
		int rshift, gshift, bshift;
		
		/* shift r,g,b so that high bit of 16-bit color specification
		 * is aligned with high bit of r,g,b-mask in visual, 
		 * AND each component with its mask,
		 * and OR the three components together
		 */
		r = cdef->red;  g = cdef->green;  b = cdef->blue;
		rmask = visual->red_mask;
		gmask = visual->green_mask;
		bmask = visual->blue_mask;
		rshift = 15 - highbit(rmask);
		gshift = 15 - highbit(gmask);
		bshift = 15 - highbit(bmask);
		
		/* shift the bits around */
		
		r = ((rshift<0) ? (r << (-rshift)) : (r >> rshift)) & rmask;
		g = ((gshift<0) ? (g << (-gshift)) : (g >> gshift)) & gmask;
		b = ((bshift<0) ? (b << (-bshift)) : (b >> bshift)) & bmask;
		cdef->pixel = r | g | b;
		return 1;
	    }
	else
		return XAllocColor(XtDisplay(w), cm, cdef);
    }

static void FreeColors(w, cm, pixels, npixels, planes)
XeRasterWidget w;
Colormap cm;
unsigned long pixels[];
int npixels;
unsigned long planes;
    {
	Visual *visual = w->raster.visual;

	if (visual->class != TrueColor && visual->class != DirectColor)
		XFreeColors(XtDisplay(w), cm, pixels, npixels, planes);
    }


void XeRasterFreeColors(w)
XeRasterWidget w;
    {
	if (w->raster.cmap)
	    {
		/*
		** Widget has a private colormap. This private one has been
		** stored into core.colormap, and the original is saved into
		** raster.cmap.
		*/
		XClearWindow(XtDisplay(w),XtWindow(w));
		XFreeColormap(XtDisplay(w), w->core.colormap);
		w->core.colormap = w->raster.cmap;
		w->raster.cmap = 0;
	    }
	else if (w->raster.alloc)
	    {
		/*
		** Widget is using the original inherited colormap
		** in core.colormap.
		*/
		FreeColors(w, w->core.colormap, w->raster.alloc,
			   w->raster.num_allocated, 0L);
		XFlush(XtDisplay(w));
	    }
	if (w->raster.alloc)
	    {
		XtFree((void *)w->raster.alloc);
		w->raster.alloc = NULL;
	    }
    }
